#include <stdio.h>
#include <stdlib.h>
#include <glib.h>
#include <getopt.h>
#include <string.h>
#include "causalito.h"
#include "model.h"
#include "readmodel.h"
#include "graph.h"
#include "causalgraph.h"
#include "globals.h"

/*********************
 * Globals variables *
 *********************/

Options options;
Model model;

/**
 * main: Causalito main function.
 * @argc: number of command line arguments.
 * @argv: table of command line arguments.
 * 
 * Return value: should return EXIT_SUCCESS :-) 
 **/
int main(int argc, char *argv[])
{
  options = alloc_and_init_options();
  /*
   * Do not write any messages before "parse_cmd_line" is
   * done because we don't know yet where to write the different kind
   * of messages
   */
  parse_cmd_line(argc, argv);
  model = alloc_and_init_model();
  if (read_model()) {
    VERBMSG("Model read and initialized.\n");
  } else {
    ERRMSG("Reading and initializing model failed.\n");
    free_and_end(FAILURE);
  }
      
  compute_causal_graph();
  
  /*   print_causal_graph(); */

  free_and_end(SUCCESS);
  return EXIT_SUCCESS;
}

/**
 * parse_cmd_line: Parse options given on command line
 * @ac: number of command line arguments.
 * @av: table of command line arguments.
 * 
 * Return value: 
 **/
int parse_cmd_line(int ac, char *av[])
{
  int c;
  FILE * desc;
  static struct option long_options[] =
    {
      {"output-file", required_argument, NULL, 'o'},
      {"verbose", no_argument, NULL, 'v'},
      {"log-file", required_argument, NULL, 'l'},
      {"help", no_argument, NULL, 'h'},
      {"debug", no_argument, NULL, 'd'},
      {"force", no_argument, NULL, 'f'},
      {0, 0, 0, 0}
    };
  
  /*   if (ac == 1) { */
  /*     usage(); */
  /*     free_and_end(SUCCESS); */
  /*   } */
  while ((c = getopt_long(ac, av, "o:vl:hf?d",
			  long_options, NULL)) != EOF) {
    switch (c) {
    case 'o':
      options->outfile = strdup(optarg);
      break;
    case 'v':
      options->verb = True;
      break;
    case 'l':
      options->logfile = strdup(optarg);
      break;
    case 'h':
      usage();
      free_and_end(SUCCESS);
      break;
    case 'f':
      options->force = True;
      break;
    case 'd':
      options->debug = True;
      break;
    case '?':
      usage();
      free_and_end(SUCCESS);
      break;
    }
  }
  /*
   * Check that the following part of command line should be one and
   * only one file name containing the model.
   */
  if (optind < (ac - 1)) {
    ERRMSG("Too many remaining arguments:\n");
    while (optind < ac)
      ERRMSG("*** %s\n", av[optind++]);
    ERRMSG("Please give one and only one model filename\n");
    free_and_end(FAILURE);
  } else if (optind == (ac - 1))
    options->infile = strdup(av[optind]);
  /*
   * All arguments on command line have been read. Now check them
   */

  VERBMSG("Parsing command line...\n");

  if (options->infile) {
    VERBMSG("Model will be read from input file '%s'.\n", options->infile);
  }
  else
    VERBMSG("Model will be read from standard input.\n");
  if (options->outfile) {
    VERBMSG("Results will be written to output file '%s'.\n",
	    options->outfile);
  }
  else
    VERBMSG("Results will be written to standard output.\n");
  if (options->logfile) {
    VERBMSG("Messages will be written to log file '%s'.\n", options->logfile);
  }
  else
    VERBMSG("Messages will be written to standard error.\n");
  /*
   * Check that input and output/log files are not the same
   */
  if (options->infile && options->outfile
      && (!strcmp(options->infile, options->outfile))) {
      ERRMSG("Input and Output files have the same name.\n");
      free_and_end(FAILURE);
  }
  if (options->infile && options->logfile
      && (!strcmp(options->infile, options->logfile))) {
      ERRMSG("Input and Log files have the same name.\n");
      free_and_end(FAILURE);
  }
  /*
   * Check log file
   */
  if (options->logfile) {
    if (!options->force) { /* force not set */
      desc = fopen(options->logfile, "r"); /* try to open file */
      if (desc) { /* file exist! */
	ERRMSG("File '%s' already exists.\n", options->logfile);
	ERRMSG("Use --force option to overwrite it or choose an other file name.\n");
	fclose(desc);
	free_and_end(FAILURE);
      }
    }
    /* At this point, force is set or file does not exist */
    if ((desc = fopen(options->logfile, "w"))) {
      /* OK: file is writable and opened*/
      VERBMSG("From now, messages go to log file.\n");
      options->logdesc = desc;
    } else {
      ERRMSG("Unable to open file '%s' in write mode.\n", options->logfile);
      free_and_end(FAILURE);
    }
  }
  /*
   * Check input file containing the model
   */
  if (options->infile) { /* A file name is given on command line */
    desc = fopen(options->infile, "r");
    if (!desc) {
      ERRMSG("Unable to read file '%s'\n", options->infile);
      free_and_end(FAILURE);
    } else
      options->indesc = desc;
  }
  /*
   * Check output file (same procedure as for log file)
   */
  if (options->outfile) {
    if (!options->force) { /* force not set */
      desc = fopen(options->outfile, "r"); /* try to open file */
      if (desc) { /* file exist! */
	ERRMSG("File '%s' already exists.\n", options->outfile);
	ERRMSG("Use --force option to overwrite it\n");
	ERRMSG("or choose an other file name.\n");
	fclose(desc);
	free_and_end(FAILURE);
      }
    }
    /* At this point, force is set or file does not exist */
    if ((desc = fopen(options->outfile, "w"))) {
      /* OK: file is writable and opened*/
      options->outdesc = desc;
    } else {
      ERRMSG("Unable to open file '%s' in write mode.\n", options->outfile);
      free_and_end(FAILURE);
    }
  }
  /*
   * If there is a log file different from standard error, print again
   * all the file names (Useful to debug)
   */
  if (options->logfile) {
    VERBMSG("Parsing command line...\n");
    if (options->infile) {
      VERBMSG("Model will be read from input file '%s'.\n", options->infile);
    }
    else
      VERBMSG("Model will be read from standard input.\n");
    if (options->outfile) {
      VERBMSG("Results will be written to output file '%s'.\n",
	      options->outfile);
    }
    else
      VERBMSG("Results will be written to standard output.\n");
    VERBMSG("Messages will be written to log file '%s'.\n", options->logfile);
  }
  VERBMSG("Parsing command line done.\n");
  return SUCCESS;
}

/**
 * alloc_and_init_options: initialize a new fresh options table.
 * @void: 
 * 
 * Return value: the options table.
 **/
Options alloc_and_init_options(void)
{
  Options o;

  o = (Options) malloc(sizeof(struct _Options));
  o->infile = NULL;
  o->outfile = NULL;
  o->logfile = NULL;
  o->indesc = stdin;
  o->outdesc = stdout;
  o->logdesc = stderr;
  o->verb = False;
  o->force = False;
  o->debug = False;
  return o;
}

/**
 * free_and_end:
 * @status: 
 * 
 * 
 * 
 * Return value: 
 **/
int free_and_end(unsigned short int status)
{
  /* dsallocation */
  
  /* Close opened files */
  if (options->outdesc != stdout)
    fclose(options->outdesc);
  if (options->indesc != stdin)
    fclose(options->indesc);
  if (options->logdesc != stderr)
    fclose(options->logdesc);
  /* Print a status message and exit */
  if (status) {
    fprintf(stderr, "Causalito ended!\n");
    exit(EXIT_SUCCESS);
  }
  else {
    fprintf(stderr, "Causalito failed!\n");
    exit(EXIT_FAILURE);
  }
  return SUCCESS;
}

/**
 * usage:
 * @void: 
 * 
 * 
 **/
void usage(void)/*Display Causalito usage and options */
{
  OUTMSG("Causalito version %s\n", VERSION);
  OUTMSG("usage: causalito [options] [file]\n");
  OUTMSG("  -v  --verbose              make causalito produce more messages\n");
  OUTMSG("  -o  --output-file <file>   send results to <file>\n");
  OUTMSG("  -l  --log-file <file>      send messages to <file>\n");
  OUTMSG("  -f  --force                allow overwriting existing files\n");
  OUTMSG("  -d  --debug                display debugging informations\n");
  OUTMSG("  -h  -? --help              display version number and this help, then exit\n");
}

/*** Local Variables: ***/
/*** change-log-default-name: "../ChangeLog" ***/
/*** End: ***/
